// ============================================================================
// Project:	A Tiny DBMS (Exercise for Course "DBMS-Implementation")
// Filename:	buf.h
// Purpose:	Buffer Manager
// Last Change:	01.02.2012
// Language:	C++
// Author:	Stefan Brass
// EMail:	brass@informatik.uni-halle.de
// WWW:		http://www.informatik.uni-halle.de/~brass/
// Address:	Feldschloesschen 15, D-06120 Halle (Saale), GERMANY
// Copyright:	(c) 2011-2012 by Stefan Brass
// Note:	There is no warranty at all - this code may contain bugs.
// ============================================================================


// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.


//=============================================================================
// Include File Frame:
//=============================================================================

#ifndef	BUF_INCLUDED
#define BUF_INCLUDED

//=============================================================================
// Used Types and Macros:
//=============================================================================

#ifndef VER_INCLUDED
#include "ver.h"
#endif

#ifndef CHECK_INCLUDED
#include "check.h"
#endif

#ifndef ERR_INCLUDED
#include "err.h"
#endif

#ifndef BLOCK_INCLUDED
#include "block.h"
#endif

// #ifndef RDBLK_INCLUDED
// #include "rdblk.h"
// #endif

#ifndef FILE_INCLUDED
#include "file.h"
#endif


//=============================================================================
// Private Constants:
//=============================================================================

//-----------------------------------------------------------------------------
// BUF_MAGIC: Magic number (identifies objects of this class).
//-----------------------------------------------------------------------------

static const long BUF_MAGIC = 0x4255461AL;	// 'BUF\n'

//-----------------------------------------------------------------------------
// BUF_MAX_SKIP: Number of modified blocks to skip in LRU list before saving.
//-----------------------------------------------------------------------------

static const int BUF_MAX_SKIP = 3;

//=============================================================================
// Class for Buffers for Blocks:
//=============================================================================

class buf_c {
	public:

//-----------------------------------------------------------------------------
// Constructor, Destructor:
//-----------------------------------------------------------------------------

	// Constructor and Destructor of this class are private.
	// See aucillary functions below.
	// One can get an object of this class only via pin().

//-----------------------------------------------------------------------------
// Class Methods (Static Member Functions):
//-----------------------------------------------------------------------------

	// Initialize this module:
	static void init();

	// Get a buffer containing a given block:
	static class buf_c *pin(file_t file, int block_no);

	// Perform checkpoint, write all modified blocks back to file:
	static bool checkpoint();

	// Return total number of buffer hits (for performance evaluation):
	static inline long int hits()
		{ return(Hits); }

	// Return total number of buffer misses (for performance evaluation):
	static inline long int misses()
		{ return(Searches - Hits); }

//-----------------------------------------------------------------------------
// (Object) Methods:
//-----------------------------------------------------------------------------

	// contents: Get the block in the buffer:
	inline block_t contents() const
		{ CHECK_VALID("buf_c::contents");
		  return(Contents); }

	// data_block: Get the block of subclass "relation data" in the buffer:
	// inline rdblk_t data_block() const
	//	{ CHECK_VALID("buf_c::data_block");
	//	  if(Contents->btype() != BTYPE_REL_DATA) {
	//		err_c::no_data_block(File->filename(), BlockNo);
	//		return 0;
	//	  }
	//	  return(static_cast<rdblk_t>(Contents)); }

	// unpin: Permit reuse of this buffer (block no longer locked to buf):
	void unpin();

	// set_modified: Mark block in buffer as modified. Must be written back
	void set_modified();

//-----------------------------------------------------------------------------
// Debugging Support:
//-----------------------------------------------------------------------------

#ifdef	VER_DEBUG
	// Integrity check:
	public:
	str_t	invalid() const;

	// Magic number (identifies objects of this class for debugging).
	private:
	long	Magic;		// Must be "BUF_MAGIC".
#endif

//-----------------------------------------------------------------------------
// Copy-constructor and assignment operator are not supported for this class:
//-----------------------------------------------------------------------------

	private:

	buf_c(const buf_c& buf);		// Not implemented
	buf_c& operator=(const buf_c& buf);	// Not implemented

//-----------------------------------------------------------------------------
// Class Attributes (Private Static Members):
//-----------------------------------------------------------------------------

	private:
// XXX
public:

	// Table of buffer objects:
	static class buf_c Bufs[VER_MAX_BUFS];
	// static class buf_c *Bufs;

	// Number of allocated buffers:
	static int NumBufs;

	// Hash Table (for searching by file and block number):
	static class buf_c *HashTab[VER_BUF_HASHSIZE];

	// LRU list (possible victims if a buffer is needed):
	static class buf_c *LRUFront;
	static class buf_c *LRURear;

	// Write list (buffers that contain modified versions of blocks):
	static class buf_c *WriteList;

	// Number of calls to pin(), this works like a time unit:
	static long int Searches;

	// Number of buffer hits (successful searches for block in buffer):
	static long int Hits;

//-----------------------------------------------------------------------------
// Attributes (Private Members):
//-----------------------------------------------------------------------------

	private:
// XXX
public:

	// Pointer to memory for a block:
	block_t Contents;

	// File to which the buffered block belongs (Null if buffer empty):
	file_t File;

	// Block number of block in buffer:
	int BlockNo;

	// Number of pins that currently lock the block to the buffer:
	int Pins;

	// Is the block modified (must be saved before buffer can be reused)?
	bool Modified;

	// Next buffer in LRU chain:
	class buf_c *LRUNext;

	// Previous buffer in LRU chain:
	class buf_c *LRUPrev;

	// Next buffer in hash chain:
	class buf_c *HashNext;

	// Backlink in hash chain (pointer to pointer that points to this):
	class buf_c **HashBacklink;

	// Next buffer in list of buffers that must be saved (written back):
	class buf_c *WriteNext;

	// Backlink for list of buffers that need saving (pointer to this):
	class buf_c **WriteBacklink;

//-----------------------------------------------------------------------------
// Auxillary Functions:
//-----------------------------------------------------------------------------

	private:

	// Constructor:
	buf_c();

	// Destructor:
	~buf_c();


	// Remove from LRU List:
	inline void lru_remove()
	{
		// The buffer must be unpinned:
		CHECK(Pins == 0, "buf_c::lru_remove: Buffer is pinned");

		// LRUFront and LRURear cannot be null (queue is nonempty):
		CHECK(LRUFront != 0, "buf_c::lru_remove: LRUFront is null");
		CHECK(LRURear  != 0, "buf_c::lru_remove: LRURear is null");

		// Adjust pointer in next element of queue (nearer to front):
		if(LRUNext == 0)
			LRUFront = LRUPrev;
		else
			LRUNext->LRUPrev = LRUPrev;

		// Adjust pointer in previous queue element (nearer to rear):
		if(LRUPrev == 0)
			LRURear = LRUNext;
		else
			LRUPrev->LRUNext = LRUNext;

		// Make the links null. Not important, but looks better:
		LRUPrev = 0;
		LRUNext = 0;
	}

	// Save block in buffer (write back to file):
	bool save_block();

//-----------------------------------------------------------------------------
// End of Class:
//-----------------------------------------------------------------------------

};

//-----------------------------------------------------------------------------
// Define pointer type:
//-----------------------------------------------------------------------------

typedef buf_c *buf_t;

//=============================================================================
// End of Include File:
//=============================================================================

#endif

